# %%
import numpy as  np 
import matplotlib.pyplot as plt 
import deerlab as dl 
from multispin_analysis_module import *
from scipy.interpolate import interp2d

# %%

def get_corrs(file):
    if 'txt' in file:
        HSCdata = np.loadtxt(file)
    if 'mat' in file:
        HSCdata = loadmat(file)['rlist']
    corrs = []
    for n1,n2 in [[0,1],[1,2],[2,0]]:
        hist,bins1,bins2 = np.histogram2d(HSCdata[:,n1],HSCdata[:,n2],bins=30)
        bins1 = bins1[:-1] + (bins1[1]-bins1[0])/2 
        bins2 = bins2[:-1] + (bins2[1]-bins2[0])/2
        xx, yy = np.meshgrid(bins1, bins2)
        grid2d = np.empty(xx.shape + (2,))
        grid2d[:, :, 0] = xx
        grid2d[:, :, 1] = yy
        
        def Pmodelflat(mean1,mean2,std1,std2,corr):
            L = np.atleast_2d([
                [std1,    0],
                [corr, std2]])
            P = Pbivar([mean1,mean2],L@L.T).pdf(grid2d)
            return P.flatten() 
        model = dl.Model(Pmodelflat)
        model.mean1.set(par0=3,lb=0,ub=8)
        model.mean2.set(par0=3,lb=0,ub=8)
        model.std1.set(par0=0.2,lb=0.01,ub=1)
        model.std2.set(par0=0.2,lb=0.01,ub=1) 
        model.corr.set(par0=0,lb=-0.5,ub=0.5)

        fit = dl.fit(model,hist.flatten())
        corrs.append(fit.corr) 
    return corrs

# %% [markdown]
# ## Model (Three spins)

# %%
#------------------------------------------------------------------------------
def threespin_dipolarsignal(t1,t2,means,cholfactors,tau1,tau2,tau3,
            lamu,lam1,lam2,lam3,lam4,lam5,lam6,lam7,lam8,lam9,lam10,lam11,lam12,lam13,lam14,lam15,lam16,lam17,lam18,lam19,lam20,lam21,lam22,lam23,lam24,
            conc_decay,d,threespin=False,fractal=False):

    tref1s,δ1s,tref2s,δ2s,λs= [],[],[],[],[]

    # Pairwise pathways modulated along t1
    reftimes_t1 = [tau1,  tau1+tau2, tau1-tau2,    0,  tau2,  tau2+tau3+tau1, -tau3, tau2+tau3]
    lams_t1     = [lam1,       lam3,      lam5, lam7,  lam9,           lam11, lam13,     lam15]

    # Pairwise pathways modulated along t2
    reftimes_t2 = [tau3, tau2+tau3,    0, tau3-tau1, tau2+tau3+tau1,  tau2, -tau1, tau1+tau2]
    lams_t2     = [lam2,      lam4, lam6,      lam8,          lam10, lam12, lam14,     lam16]

    # Pairwise pathways modulated along t2
    reftimes_t1t2 = [[-tau3,tau3], [tau2,tau2], [0, 0], [tau2+tau3+tau1,tau2+tau3+tau1], [tau2+tau3,tau2+tau3], [tau1,-tau1], [tau1+tau2,tau1+tau2], [tau1-tau3,tau3-tau1]]
    lams_t1t2     = [       lam17,       lam18,  lam19,                           lam20,                 lam21,        lam22,                 lam23,                 lam24]

    Npermutations = 6

    # Two-spin contributions
    for reftime,lam in zip(reftimes_t1,lams_t1): 
        # Pathway set 1 - two-spin, t1-modulated
        tref1, tref2 = [ reftime, None, None], [None, None, None]
        δ1, δ2       = [       1,    0,    0], [   0,    0,    0]
        tref1s, tref2s = [add_permutations(trefs,tref) for trefs,tref in zip([tref1s,tref2s],[tref1,tref2])]
        δ1s, δ2s = [add_permutations(δs,δ) for δs,δ in zip([δ1s,δ2s],[δ1,δ2])]
        λs += [lam*lamu*lamu]*Npermutations

    for reftime,lam in zip(reftimes_t2,lams_t2): 
        # Pathway set 1 - two-spin, t1-modulated
        tref1, tref2 = [ None, None, None], [ reftime, None, None]
        δ1, δ2       = [    0,    0,    0], [       1,    0,    0]
        tref1s, tref2s = [add_permutations(trefs,tref) for trefs,tref in zip([tref1s,tref2s],[tref1,tref2])]
        δ1s, δ2s = [add_permutations(δs,δ) for δs,δ in zip([δ1s,δ2s],[δ1,δ2])]
        λs += [lam*lamu*lamu]*Npermutations

    for reftime,lam in zip(reftimes_t1t2,lams_t1t2): 
        if lam>0.001:
            # Pathway set 1 - two-spin, t1-modulated
            tref1, tref2 = [ reftime[0], None, None], [ reftime[1], None, None]
            δ1, δ2       = [          1,    0,    0], [          1,    0,    0]
            tref1s, tref2s = [add_permutations(trefs,tref) for trefs,tref in zip([tref1s,tref2s],[tref1,tref2])]
            δ1s, δ2s = [add_permutations(δs,δ) for δs,δ in zip([δ1s,δ2s],[δ1,δ2])]
            λs += [lam*lamu*lamu]*Npermutations

    ### Three-spin contributions
    if threespin:
        for reftime1,lam1 in zip(reftimes_t1,lams_t1): 
            for reftime2,lam2 in zip(reftimes_t1,lams_t1): 
                if lam1*lam2*lamu>0.001:
                    # Pathway set 7 - three-spin, t1,t2-modulated
                    tref1, tref2 = [ reftime1, reftime2, None], [None, None, None]
                    δ1, δ2       = [        1,        1,    0], [   0,    0,    0]
                    tref1s, tref2s = [add_permutations(trefs,tref) for trefs,tref in zip([tref1s,tref2s],[tref1,tref2])]
                    δ1s, δ2s = [add_permutations(δs,δ) for δs,δ in zip([δ1s,δ2s],[δ1,δ2])]
                    λs += [lam1*lam2*lamu]*Npermutations
        for reftime1,lam1 in zip(reftimes_t2,lams_t2): 
            for reftime2,lam2 in zip(reftimes_t2,lams_t2): 
                if lam1*lam2*lamu>0.001:
                    # Pathway set 7 - three-spin, t1,t2-modulated
                    tref1, tref2 = [ None, None, None], [ reftime1, reftime2, None]
                    δ1, δ2       = [    0,    0,    0], [        1,        1,    0]
                    tref1s, tref2s = [add_permutations(trefs,tref) for trefs,tref in zip([tref1s,tref2s],[tref1,tref2])]
                    δ1s, δ2s = [add_permutations(δs,δ) for δs,δ in zip([δ1s,δ2s],[δ1,δ2])]
                    λs += [lam1*lam2*lamu]*Npermutations
        for reftime1,lam1 in zip(reftimes_t1,lams_t1): 
            for reftime2,lam2 in zip(reftimes_t2,lams_t2): 
                if lam1*lam2*lamu>0.001:
                    # Pathway set 7 - three-spin, t1,t2-modulated
                    tref1, tref2 = [ reftime1, None, None], [ None, reftime2, None]
                    δ1, δ2       = [        1,    0,    0], [    0,        1,    0]
                    tref1s, tref2s = [add_permutations(trefs,tref) for trefs,tref in zip([tref1s,tref2s],[tref1,tref2])]
                    δ1s, δ2s = [add_permutations(δs,δ) for δs,δ in zip([δ1s,δ2s],[δ1,δ2])]
                    λs += [lam1*lam2*lamu]*Npermutations

    # Monte-Carlo multivariate distance integration
    Nsamples = 500000
    np.random.seed(seed=1)
    rsamples = Pmultivar(means,cholfactors).rvs(Nsamples)
    rsamples = np.maximum(rsamples,1e-16) # Avoid values exactly at zero 

    # Triangle inequalities
    Ndistances = np.shape(rsamples)[1]
    idx = np.arange(Ndistances)
    triangle_condition = np.full_like(rsamples,False)
    for n in range(Ndistances):
        idx = np.roll(idx,-1,axis=0)
        triangle_condition[:,n] = np.sum(rsamples[:,idx[:-1]],axis=1) > rsamples[:,idx[-1]]
    triangle_condition = np.all(triangle_condition,axis=1)
    # Discard samples that do not satisfy triangle inequalities
    rsamples = rsamples[triangle_condition,:]

    # Construct the dipolar signal 
    λ0 = np.maximum(1 - np.sum(λs),0)
    Vintra,Vinter = λ0,1
    twospin_contributions,threespin_contributions = [],[]
    # Loop over all dipolar pathways
    for δ1,tref1,δ2,tref2,λ in zip(δ1s,tref1s,δ2s,tref2s,λs):
        
        # Set trefs defined as None to an arbitrary numerical value
        tref1,tref2 = [[tref[n] if δn!=0 else 0 for n,δn in enumerate(δ)] for tref,δ in zip([tref1,tref2],[δ1,δ2])]

        # Number of spins participating in the pathway
        Nspin = np.sum(δ1)+np.sum(δ2)+1 
                    
        # Two-spin intramolecular contributions
        if Nspin==2:
            Vintra_contr = 0
            if np.sum(δ1)>0:
                n = int(np.where(δ1==1)[0])

                # Estimate the 1D-marginal distance distribution
                Pmarginal,bins = np.histogram(rsamples[:,n], bins=300)
                rgrid = (bins[:-1] + bins[1:])/2
                Pmarginal = Pmarginal/np.trapz(Pmarginal,rgrid)

                # Two-spin intramolecular contribution modulated along t1-dimension
                Vintra_contr_ = dl.dipolarkernel(δ1[n]*(t1-tref1[n]),rgrid)@Pmarginal
                Vintra_contr += np.repeat(Vintra_contr_[:,np.newaxis] ,len(t2),axis=1)

            if np.sum(δ2)>0:
                n = int(np.where(δ2==1)[0])

                # Estimate the 1D-marginal distance distribution
                Pmarginal,bins = np.histogram(rsamples[:,n], bins=300)
                rgrid = (bins[:-1] + bins[1:])/2
                Pmarginal = Pmarginal/np.trapz(Pmarginal,rgrid)

                # Two-spin intramolecular contribution modulated along t2-dimension
                Vintra_contr_ = dl.dipolarkernel(δ2[n]*(t2-tref2[n]),rgrid)@Pmarginal
                Vintra_contr += np.repeat(Vintra_contr_[np.newaxis,:] ,len(t1),axis=0)
                
            twospin_contributions.append(λ*Vintra_contr)


        # Three-spin intramolecular contributions
        if Nspin==3:

            # Non-uniformly subsample the time-domain (speedup)
            n1 = int(np.where(δ1==1)[0][0]) if np.sum(δ1)>0 else 0
            n2 = int(np.where(δ2==1)[0][0]) if np.sum(δ2)>0 else 0
            np.random.seed(0)
            subsampling1 = np.exp(-6*abs(t1-tref1[n1])/max(t1))*np.random.rand(len(t1))>0.025
            subsampling2 = np.exp(-6*abs(t2-tref2[n2])/max(t2))*np.random.rand(len(t2))>0.025
            subsampling1[-1] = 1
            subsampling2[-1] = 1
            t1sub = t1[subsampling1]
            t2sub = t2[subsampling2]

            # Average three-spin contribution over all distances
            Nsamples = 100
            Vintra_contr = 0
            for rs in rsamples[:Nsamples,:]: 
                r1,r2,r3 = rs
                # Three-spin intramolecular contribution
                Vintra_contr += threespin_dipolarkernel2D(t1sub,t2sub,r1,r2,r3,δ1,tref1,δ2,tref2) 
            Vintra_contr /= Nsamples
            Vintra_contr = interp2d(t1sub,t2sub,Vintra_contr.T)(t1,t2).T
            threespin_contributions.append(λ*Vintra_contr)

        # Incorporate intermolecular contribution
        Vintra += λ*Vintra_contr

        # Intermolecular contribution
        Vinter *= Bbasis2D(t1,t2,λ,conc_decay,d,δ1,δ2,tref1,tref2,fractal=fractal)

    Vinter = Vinter/np.max(Vinter)
    
    # Construct the total dipolar signal
    V = Vintra*Vinter
    V = np.maximum(V,1e-16)

    return V,λ0,Vinter,Vintra,twospin_contributions,threespin_contributions
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
def construct_threespin_dipolarmodel(t1,t2,tau1_exp,tau2_exp,tau3_exp,threespin=False,fractal=False):

    def _threespin_dipolarsignal(mean1,mean2,mean3,chol11,chol22,chol33,chol21,chol31,chol32,tau1,tau2,tau3,
                            lamu,lam1,lam2,lam3,lam4,lam5,lam6,lam7,lam8,lam9,lam10,lam11,lam12,lam13,lam14,lam15,
                        lam16,lam17,lam18,lam19,lam20,lam21,lam22,lam23,lam24,conc_decay,d):
            Vfit = threespin_dipolarsignal(t1,t2,[mean1,mean2,mean3],[chol11,chol22,chol33,chol21,chol31,chol32],tau1,tau2,tau3,
                                    lamu,lam1,lam2,lam3,lam4,lam5,lam6,lam7,lam8,lam9,lam10,lam11,lam12,lam13,lam14,lam15,
                            lam16,lam17,lam18,lam19,lam20,lam21,lam22,lam23,lam24,conc_decay,d,threespin=threespin, fractal=fractal)[0]
            Vfit = np.atleast_2d(Vfit.flatten()).T # Flatten to 1D-vector
            return Vfit 
    # Construct model 
    Vmultispin = dl.Model(_threespin_dipolarsignal)
    Vmultispin.description = 'Multi-spin dipolar signal arising from a three-spin TRIER experiment'
    # Set parameter properties
    Vmultispin.addlinear('scale',lb=0, description='Echo amplitude in the abscense of dipolar modulations')
    Vmultispin.mean1.set(lb=1.0, ub=8.0, par0=3.0, unit='nm', description='Average inter-spin distance  #1')
    Vmultispin.mean2.set(lb=1.0, ub=8.0, par0=3.0, unit='nm', description='Average inter-spin distance  #2')
    Vmultispin.mean3.set(lb=1.0, ub=8.0, par0=3.0, unit='nm', description='Average inter-spin distance  #3')
    Vmultispin.chol11.set(lb=0.0, ub=0.8, par0=0.40, unit='nm', description='Cholesky factor ℓ11' )
    Vmultispin.chol22.set(lb=0.0, ub=0.8, par0=0.40, unit='nm', description='Cholesky factor ℓ22' )
    Vmultispin.chol33.set(lb=0.0, ub=0.8, par0=0.40, unit='nm', description='Cholesky factor ℓ33' )
    Vmultispin.chol21.set(lb=-1.0, ub=1.0, par0=0.0, unit='nm', description='Cholesky factor ℓ21' )
    Vmultispin.chol31.set(lb=-1.0, ub=1.0, par0=0.0, unit='nm', description='Cholesky factor ℓ31' )
    Vmultispin.chol32.set(lb=-1.0, ub=1.0, par0=0.0, unit='nm', description='Cholesky factor ℓ32' )
    Vmultispin.tau1.set(lb=tau1_exp-0.2, ub=tau1_exp+0.2, par0=tau1_exp, unit='μs', description='First inter-pulse delay' )
    Vmultispin.tau2.set(lb=tau2_exp-0.2, ub=tau2_exp+0.2, par0=tau2_exp, unit='μs', description='Second inter-pulse delay' )
    Vmultispin.tau3.set(lb=tau3_exp-0.2, ub=tau3_exp+0.2, par0=tau3_exp, unit='μs', description='Third inter-pulse delay' )
    Vmultispin.lamu.set(lb=0.00, ub=2.00, par0=1.00, unit=None, description='Amplitude of unmodulated pairwise pathway' )
    Vmultispin.lam1.set(lb=0.00, ub=1.00, par0=0.02, unit=None, description='Dipolar pathway #1 probability' )
    Vmultispin.lam2.set(lb=0.00, ub=1.00, par0=0.02, unit=None, description='Dipolar pathway #2 probability' )
    Vmultispin.lam3.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #3 probability' )
    Vmultispin.lam4.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #4 probability' )
    Vmultispin.lam5.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #5 probability' )
    Vmultispin.lam6.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #6 probability' )
    Vmultispin.lam7.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #7 probability' )
    Vmultispin.lam8.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #8 probability' )
    Vmultispin.lam9.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #9 probability' )
    Vmultispin.lam10.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #10 probability' )
    Vmultispin.lam11.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #11 probability' )
    Vmultispin.lam12.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #12 probability' )
    Vmultispin.lam13.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #13 probability' )
    Vmultispin.lam14.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #14 probability' )
    Vmultispin.lam15.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #15 probability' )
    Vmultispin.lam16.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #16 probability' )
    Vmultispin.lam17.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #17 probability' )
    Vmultispin.lam18.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #18 probability' )
    Vmultispin.lam19.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #19 probability' )
    Vmultispin.lam20.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #20 probability' )
    Vmultispin.lam21.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #21 probability' )
    Vmultispin.lam22.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #22 probability' )
    Vmultispin.lam23.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #23 probability' )
    Vmultispin.lam24.set(lb=0.00, ub=1.00, par0=0.00, unit=None, description='Dipolar pathway #24 probability' )
    Vmultispin.conc_decay.set(lb=0, ub=50, par0=0.005, unit='μM', description='Spin concentration / Decay rate' )
    Vmultispin.d.set(lb=1, ub=3, par0=1.20, unit='μM', description='Stretch factor' )
    if not fractal:
        Vmultispin.d.freeze(1)
        Vmultispin.conc_decay.set(lb=0, ub=500, par0=40, unit='μM', description='Spin concentration / Decay rate' )
    else:
        Vmultispin.conc_decay.set(lb=0, ub=5, par0=0.05, unit='μM', description='Spin concentration / Decay rate' )

    return Vmultispin
#------------------------------------------------------------------------------

# %% [markdown]
# ----------------
# ## Triradical T111
# ----------------

# %%
threespin = True 
fractal = True

# Load experimental data
data = loadmat(r'.\\data\\triradical_T111_TRIER.mat')
HSCfile=r'.\\HSC_simulations\\HSCsim_T111_100000samples.txt'
t1 = np.squeeze(data['t1'])/1000 # ns -> μs
t2 = np.squeeze(data['t2'])/1000 # ns -> μs
Vexp = np.squeeze(data['V2d']).T
tau1_exp = np.squeeze(data['tau1'])/1000 # ns -> μs
tau2_exp = np.squeeze(data['tau2'])/1000 # ns -> μs
tau3_exp = np.squeeze(data['tau3'])/1000 # ns -> μs
Vexp /= np.max(Vexp) 

# Subsample the signal due to limited memory
subsampling1 = np.arange(0,len(t1),3)
subsampling2 = np.arange(0,len(t2),3)
t1_sub = t1[subsampling1]
t2_sub = t2[subsampling2]
Vexp_sub = Vexp[np.ix_(subsampling1,subsampling2)]

# Construct the three-spin TRIER model
Vmodel_TRIER = construct_threespin_dipolarmodel(t1_sub,t2_sub,tau1_exp,tau2_exp,tau3_exp,threespin=threespin,fractal=fractal)

# Use the HSC simulation to get a good starting point
HSCdata = np.loadtxt(HSCfile)
for n in range(HSCdata.shape[1]):
    hist,bins = np.histogram(HSCdata[:,n],bins=200)
    bins = bins[:-1] + (bins[1]-bins[0])/2 
    fit = dl.fit(dl.dd_gauss,hist,bins)
    getattr(Vmodel_TRIER,f'mean{n+1}').par0 = fit.mean
    getattr(Vmodel_TRIER,f'chol{n+1}{n+1}').par0 = fit.std
corrs = get_corrs(HSCfile)
Vmodel_TRIER.chol21.par0 = corrs[0]
Vmodel_TRIER.chol31.par0 = corrs[1]
Vmodel_TRIER.chol32.par0 = corrs[2]

# Ignore pathways refocusing outside of the detected signal
Vmodel_TRIER.lam5.freeze(0)
Vmodel_TRIER.lam6.freeze(0)
Vmodel_TRIER.lam7.freeze(0)
Vmodel_TRIER.lam8.freeze(0)
Vmodel_TRIER.lam13.freeze(0)
Vmodel_TRIER.lam14.freeze(0)
Vmodel_TRIER.lam17.freeze(0)
Vmodel_TRIER.lam18.freeze(0)
Vmodel_TRIER.lam19.freeze(0)
Vmodel_TRIER.lam20.freeze(0)
Vmodel_TRIER.lam21.freeze(0)
Vmodel_TRIER.lam22.freeze(0)
Vmodel_TRIER.lam23.freeze(0)
Vmodel_TRIER.lam24.freeze(0)

# Fit the model to the data 
results = dl.fit(Vmodel_TRIER, Vexp_sub.flatten(), reg=False, ftol=1e-3, max_nfev=400,verbose=2) 

# Print the fit summary 
print(results)

# Plot the multivariate distance distribution
means = [results.mean1, results.mean2, results.mean3]
cholesky_factors = [results.chol11,results.chol22,results.chol33,results.chol21, results.chol31, results.chol32] 
plot_multivariate(means,cholesky_factors,xlim=[0,8],saveas='TRIER_T111_Pfit.svg',HSCfile=HSCfile)

# Plot the fitted TRIER signal
params = [t1,t2,means,cholesky_factors,results.tau1,results.tau2,results.tau3,results.lamu] + [getattr(results,f'lam{n+1}') for n in range(24)] + [results.conc_decay,results.d,threespin,fractal]
plot_trier_fit(t1,t2,Vexp,threespin_dipolarsignal,params,results.scale,saveas='TRIER_T111_Vfit.svg',levels=60)

# %% [markdown]
# -----------
# ## Triradical T011
# -----------

# %%
threespin = True 
fractal = True

# Load experimental data
data = loadmat(r'.\\data\\triradical_T011_TRIER.mat')
HSCfile=r'.\\HSC_simulations\\HSCsim_triradical_T011_100000samples.txt'
t1 = np.squeeze(data['t1'])/1000 # ns -> μs
t2 = np.squeeze(data['t2'])/1000 # ns -> μs
Vexp = np.squeeze(data['V2d']).T
tau1_exp = np.squeeze(data['tau1'])/1000 # ns -> μs
tau2_exp = np.squeeze(data['tau2'])/1000 # ns -> μs
tau3_exp = np.squeeze(data['tau3'])/1000 # ns -> μs
Vexp /= np.max(Vexp) 

# Subsample the signal due to limited memory
subsampling1 = np.arange(0,len(t1),2)
subsampling2 = np.arange(0,len(t2),2)
t1_sub = t1[subsampling1]
t2_sub = t2[subsampling2]
Vexp_sub = Vexp[np.ix_(subsampling1,subsampling2)]

# Construct the three-spin TRIER model
Vmodel_TRIER = construct_threespin_dipolarmodel(t1_sub,t2_sub,tau1_exp,tau2_exp,tau3_exp,threespin=threespin,fractal=fractal)

# Use the HSC simulation to get a good starting point
HSCdata = np.loadtxt(HSCfile)
for n in range(HSCdata.shape[1]):
    hist,bins = np.histogram(HSCdata[:,n],bins=200)
    bins = bins[:-1] + (bins[1]-bins[0])/2 
    fit = dl.fit(dl.dd_gauss,hist,bins)
    getattr(Vmodel_TRIER,f'mean{n+1}').par0 = fit.mean
    getattr(Vmodel_TRIER,f'chol{n+1}{n+1}').par0 = fit.std
corrs = get_corrs(HSCfile)
Vmodel_TRIER.chol21.par0 = corrs[0]
Vmodel_TRIER.chol31.par0 = corrs[1]
Vmodel_TRIER.chol32.par0 = corrs[2]

# Ignore pathways refocusing outside of the detected signal
Vmodel_TRIER.lam5.freeze(0)
Vmodel_TRIER.lam6.freeze(0)
Vmodel_TRIER.lam7.freeze(0)
Vmodel_TRIER.lam8.freeze(0)
Vmodel_TRIER.lam13.freeze(0)
Vmodel_TRIER.lam14.freeze(0)
# Ignore two-dimensionally modulated pathways 
Vmodel_TRIER.lam17.freeze(0)
Vmodel_TRIER.lam18.freeze(0)
Vmodel_TRIER.lam19.freeze(0)
Vmodel_TRIER.lam20.freeze(0)
Vmodel_TRIER.lam21.freeze(0)
Vmodel_TRIER.lam22.freeze(0)
Vmodel_TRIER.lam23.freeze(0)
Vmodel_TRIER.lam24.freeze(0)

# Fit the model to the data 
results = dl.fit(Vmodel_TRIER, Vexp_sub.flatten(), reg=False, ftol=1e-4, max_nfev=350) 

# Print the fit summary 
print(results)

# Plot the multivariate distance distribution
means = [results.mean1, results.mean2, results.mean3]
cholesky_factors = [results.chol11,results.chol22,results.chol33,results.chol21, results.chol31, results.chol32] 
plot_multivariate(means,cholesky_factors,xlim=[0,8],saveas='TRIER_T011_Pfit.svg', HSCfile=HSCfile)

# Plot the fitted TRIER signal
params = [t1,t2,means,cholesky_factors,results.tau1,results.tau2,results.tau3, results.lamu] + [getattr(results,f'lam{n+1}') for n in range(24)] + [results.conc_decay,results.d,threespin,fractal]
plot_trier_fit(t1,t2,Vexp,threespin_dipolarsignal,params,results.scale,saveas='TRIER_T011_Vfit.svg',levels=45)

# %% [markdown]
# -----------
# ## Triradical Rpo47
# -----------

# %%
threespin = True 
fractal = False

# Load experimental data
data = loadmat('.\\data\\triradical_Rpo47_TRIER.mat')
MMMfile = r".\\MMMx_simulations\\MMMx_Rpo47_triple_labelled_sampled.mat"
t1 = np.squeeze(data['t1'])/1000 # ns -> μs
t2 = np.squeeze(data['t2'])/1000 # ns -> μs
Vexp = np.squeeze(data['V2d']).T
tau1_exp = np.squeeze(data['tau1'])/1000 # ns -> μs
tau2_exp = np.squeeze(data['tau2'])/1000 # ns -> μs
tau3_exp = np.squeeze(data['tau3'])/1000 # ns -> μs
Vexp /= np.max(Vexp) 

# Subsample the signal due to limited memory
subsampling1 = np.arange(0,len(t1),2)
subsampling2 = np.arange(0,len(t2),2)
t1_sub = t1[subsampling1]
t2_sub = t2[subsampling2]
Vexp_sub = Vexp[np.ix_(subsampling1,subsampling2)]

# Construct the three-spin TRIER model
Vmodel_TRIER = construct_threespin_dipolarmodel(t1_sub,t2_sub,tau1_exp,tau2_exp,tau3_exp,threespin=threespin,fractal=fractal)

# Use the HSC simulation to get a good starting point
from scipy.io import loadmat
MMMdata = loadmat(MMMfile)['rlist']
for n in range(MMMdata.shape[1]):
    hist,bins = np.histogram(MMMdata[:,n],bins=200)
    bins = bins[:-1] + (bins[1]-bins[0])/2 
    fit = dl.fit(dl.dd_gauss,hist,bins)
    getattr(Vmodel_TRIER,f'mean{n+1}').par0 = fit.mean
    getattr(Vmodel_TRIER,f'chol{n+1}{n+1}').par0 = fit.std
corrs = get_corrs(MMMfile)
Vmodel_TRIER.chol21.par0 = corrs[0]
Vmodel_TRIER.chol31.par0 = corrs[1]
Vmodel_TRIER.chol32.par0 = corrs[2]

# Ignore pathways refocusing outside of the detected signal
Vmodel_TRIER.lam5.freeze(0)
Vmodel_TRIER.lam6.freeze(0)
Vmodel_TRIER.lam7.freeze(0)
Vmodel_TRIER.lam8.freeze(0)
Vmodel_TRIER.lam13.freeze(0)
Vmodel_TRIER.lam14.freeze(0)

Vmodel_TRIER.lam17.freeze(0)
Vmodel_TRIER.lam18.freeze(0)
Vmodel_TRIER.lam19.freeze(0)
Vmodel_TRIER.lam20.freeze(0)
Vmodel_TRIER.lam21.freeze(0)
Vmodel_TRIER.lam22.freeze(0)
Vmodel_TRIER.lam23.freeze(0)
Vmodel_TRIER.lam24.freeze(0)

# Fit the model to the data 
results = dl.fit(Vmodel_TRIER, Vexp_sub.flatten(), reg=False, ftol=1e-4, max_nfev=300, verbose=2) 

# Print the fit summary 
print(results)

# Plot the multivariate distance distribution
means = [results.mean1, results.mean2, results.mean3]
cholesky_factors = [results.chol11,results.chol22,results.chol33,results.chol21, results.chol31, results.chol32] 
plot_multivariate(means,cholesky_factors,xlim=[0,8],saveas='TRIER_Rpo47_Pfit.svg',
        MMMfile = r"D:\lufa\projects\Multipathway  3spins\rpo47_triple_labelled_MMMx_sampled.mat")

# Plot the fitted TRIER signal
params = [t1,t2,means,cholesky_factors,results.tau1,results.tau2,results.tau3,results.lamu] + [getattr(results,f'lam{n+1}') for n in range(24)] + [results.conc_decay,results.d,threespin,fractal]
plot_trier_fit(t1,t2,Vexp,threespin_dipolarsignal,params,results.scale,saveas='TRIER_Rpo47_Vfit.svg',levels=40)

# %%



