import numpy as np 
import warnings 
import matplotlib.pyplot as plt
import deerlab as dl 
from numpy import cos, sin, sqrt, arccos
from numpy import pi as π
from scipy.stats import multivariate_normal
import scipy.stats as st
from deerlab.bg_models import _strexp,_hom3d
import numexpr as ne
import matplotlib.gridspec as gridspec
from scipy.io import loadmat
from scipy import optimize
from scipy.interpolate import interp1d
from itertools import permutations

# Fundamental constants (CODATA 2018)
ge = 2.00231930436256  # free-electron g factor
muB = 9.2740100783e-24 # Bohr magneton, J/T 
mu0 = 1.25663706212e-6 # magnetic constant, N A^-2 = T^2 m^3 J^-1 
h = 6.62607015e-34     # Planck constant, J/Hz
def w0(g):
    return (mu0/2)*muB**2*g[0]*g[1]/h*1e21 # Hz m^3 -> MHz nm^3 -> Mrad s^-1 nm^3

# ----------------------------------------------------------------------------------
def cholesky_covmat(L):

    if len(L)==6:
        L = np.atleast_2d([
                [L[0],   0,   0],
                [L[3],L[1],   0],
                [L[4],L[5],L[2]],
            ])

    elif len(L)==21:
        L = np.atleast_2d([
                [L[0],    0,     0,    0,     0,     0],
                [L[6], L[1],     0,    0,     0,     0],
                [L[7], L[11], L[2],    0,     0,     0],
                [L[8], L[12], L[15],L[3],     0,     0],
                [L[9], L[13], L[16],L[18], L[4],     0],
                [L[10],L[14], L[17],L[19], L[20], L[5]]
            ])
    else: 
        raise ValueError('The Cholesky decomposition vector has not the correct number of elements')

    L = np.atleast_2d(L)
    Σ = L@L.T
    return Σ 
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def Punivar(mean,variance):
    """
    Univariate distance distribution
    --------------------------------

    Parameters
    ----------
    mean : array_like
        Mean distance of the triangle side
    variance : matrix_like
        Variance
    
    Returns
    -------
    P : distribution object 
        Univariate normal distribution.
    """
    Σ = np.atleast_2d([[variance]])
    mean = np.atleast_1d(mean)

    # Generate the multivariate distribution
    Pvar = multivariate_normal(mean, cov=Σ)

    return Pvar
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def Pbivar(means,Σ):
    """
    Univariate distance distribution
    --------------------------------

    Parameters
    ----------
    mean : array_like
        Mean distance of the triangle side
    Σ : matrix_like
        Covariance matrix
    
    Returns
    -------
    P : distribution object 
        Univariate normal distribution.
    """
    means = np.atleast_1d(means)

    # Generate the multivariate distribution
    Pvar = multivariate_normal(means, cov=Σ)

    return Pvar
# ----------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------
def Pmultivar(means,cholfactors):
    """
    Multivariate distance distribution
    --------------------------------

    Parameters
    ----------
    means : array_like
        List of mean distances of the triangle sides
    Σ : matrix_like
        Covariance matrix
    
    Returns
    -------
    P : distribution object 
        Multivariate normal distribution.
    """

    Σ = cholesky_covmat(cholfactors)
    # Generate the multivariate distribution
    Pvar = multivariate_normal(means, cov=Σ)

    return Pvar
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def threespin_dipolarkernel(t,r1,r2,r3,δ,tref):

    t,r1,r2,r3 = np.atleast_1d(t,r1,r2,r3)
    nt = t.size
    nr = r1.size

    # Estimate RAM required to hold the kernel
    memory_req = 4/1e9*nt*500**2
    if memory_req>5: raise MemoryError(f'The requested kernel would cost {memory_req:.2f}GB. Aborted.')

    tref_1, tref_2, tref_3 = tref
    δ_1, δ_2, δ_3 = δ

    # Prepare slicing for broadcasting
    x  = np.newaxis 
    t    =    t[:,x,x]
    r1    =  r1[x,:,x]
    r2    =  r2[x,:,x]
    r3    =  r3[x,:,x]

    # Orientational grid (uniform distributions)
    data = loadmat(r'.\\data\\spherical_grid_CI_12order.mat')
    φ1 = np.squeeze(data['phi'])
    θ1 = np.squeeze(data['theta'])
    weights = np.squeeze(data['weights'])
    weights = weights/np.sum(weights)
    cosθ1 = np.cos(θ1)
    sinθ1 = np.sin(θ1)
    cosφ1 = np.cos(φ1)

    #nKnots = 300
    #θ1 = np.random.uniform(low=0,high=π/2,size=nKnots)
    #φ1 = np.random.uniform(low=0,high=2*π,size=nKnots)
    #sinθ1 = np.sin(θ1)
    #cosθ1 = np.cos(θ1)
    #cosφ1 = np.cos(φ1)
    #weights = sinθ1/np.sum(sinθ1)

    cosθ1 = cosθ1[x,x,:]
    sinθ1 = sinθ1[x,x,:]
    cosφ1 = cosφ1[x,x,:]
    weights = weights[x,x,:]
    
    # Triangle inequalities
    triangle1 = (r1 + r2 > r3)  &  (r1 + r3 > r2) & (r2 + r3 > r1)
    triangle2 = (np.abs(r1 - r2) < r3) & (np.abs(r1 - r3) < r2) & (np.abs(r2 - r3) < r1)
    triangle = np.repeat(triangle1 & triangle2, nt, axis=0)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        # Law of cosines
        Δθ12 = arccos((r1**2 + r2**2 - r3**2)/(2*r1*r2))
        Δθ13 = -arccos((r1**2 + r3**2 - r2**2)/(2*r1*r3))

        sinΔθ12, cosΔθ12 = sin(Δθ12), cos(Δθ12)
        sinΔθ13, cosΔθ13 = sin(Δθ13), cos(Δθ13)

        cosθ2 = sinΔθ12*cosφ1*sinθ1 + cosΔθ12*cosθ1
        cosθ3 = sinΔθ13*cosφ1*sinθ1 + cosΔθ13*cosθ1

        # Dipolar frequencies
        ωr1 = w0([ge,ge])/(r1**3)  # rad s^-1
        ωr2 = w0([ge,ge])/(r2**3)  # rad s^-1
        ωr3 = w0([ge,ge])/(r3**3)  # rad s^-1

        # Perform broadcast operation
        ωr1_tdip1 = ωr1*(δ_1)*(t-tref_1) 
        ωr2_tdip2 = ωr2*(δ_2)*(t-tref_2) 
        ωr3_tdip3 = ωr3*(δ_3)*(t-tref_3) 

        K = ne.evaluate('cos(ωr1_tdip1*(1 - 3*cosθ1**2) )* \
                         cos(ωr2_tdip2*(1 - 3*cosθ2**2) )* \
                         cos(ωr3_tdip3*(1 - 3*cosθ3**2) )')

        # Powder averaging 
        K = np.sum(weights*K,2)

        K = np.squeeze(K)

    return K
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def threespin_dipolarkernel2D(t1,t2,r1,r2,r3,δ1,tref1,δ2,tref2):

    t1,t2,r1,r2,r3 = np.atleast_1d(t1,t2,r1,r2,r3)
    nt1 = t1.size
    nt2 = t2.size
    nr1 = r1.size
    nr2 = r2.size  
    nr3 = r3.size    

    tref1_1, tref1_2, tref1_3 = tref1
    tref2_1, tref2_2, tref2_3 = tref2
    δ1_1, δ1_2, δ1_3 = δ1
    δ2_1, δ2_2, δ2_3 = δ2

    # Orientational grid (uniform distributions)
    data = loadmat(r'.\\data\\spherical_grid_CI_12order.mat')
    φ = np.squeeze(data['phi'])
    θ = np.squeeze(data['theta'])
    weights = np.squeeze(data['weights'])
    weights = weights/np.sum(weights)
    cosθ = np.cos(θ)
    sinθ = np.sin(θ)
    cosφ = np.cos(φ)

    # Estimate RAM required to hold the kernel
    memory_req = 4/1e9*nt1*nt2*nr1*nr2*nr3*len(φ)
    if memory_req>5: raise MemoryError(f'The requested kernel would cost {memory_req:.2f}GB. Aborted.')


    # Broadcasting strategies
    x  = np.newaxis 
    t1      =      t1[:,x,x,x]
    t2      =      t2[x,:,x,x]
    r1      =      r1[x,x,:,x]
    r2      =      r2[x,x,:,x]
    r3      =      r3[x,x,:,x]
    cosθ    =    cosθ[x,x,x,:]
    sinθ    =    sinθ[x,x,x,:]
    cosφ    =    cosφ[x,x,x,:]
    weights = weights[x,x,x,:]

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        # Law of cosines
        s = (r1+r2+r3)/2
        Δθ12 = 2*arccos(sqrt((s-r1)*(s-r2)/(r1*r2)))
        Δθ13 = 2*arccos(sqrt((s-r1)*(s-r3)/(r1*r3)))

        sinΔθ12, cosΔθ12 = sin(Δθ12), cos(Δθ12)
        sinΔθ13, cosΔθ13 = sin(Δθ13), cos(Δθ13)

        # Dipolar frequencies
        ωr1 = w0([ge,ge])/(r1**3)  # rad s^-1
        ωr2 = w0([ge,ge])/(r2**3)  # rad s^-1
        ωr3 = w0([ge,ge])/(r3**3)  # rad s^-1

        # Perform broadcast operation
        ωr1_tdip1 = ωr1*(abs(δ1_1)*(t1-tref1_1) + abs(δ2_1)*(t2-tref2_1))
        ωr2_tdip2 = ωr2*(abs(δ1_2)*(t1-tref1_2) + abs(δ2_2)*(t2-tref2_2))
        ωr3_tdip3 = ωr3*(abs(δ1_3)*(t1-tref1_3) + abs(δ2_3)*(t2-tref2_3))

        K = ne.evaluate('cos(ωr1_tdip1*(1 - 3*cosθ**2) +  \
                    ωr2_tdip2*(1 - 3*(sinΔθ12*cosφ*sinθ + cosΔθ12*cosθ)**2) + \
                        ωr3_tdip3*(1 - 3*(sinΔθ13*cosφ*sinθ + cosΔθ13*cosθ)**2))')

    # Powder averaging 
    K = np.sum(weights*K,-1)
    K = np.squeeze(K)

    return K
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def Bbasis(t,λ,conc_decay,d,δ,tref,fractal=False):

    Bbasis = 1
    for δ_q,tref_q in zip(δ,tref):
        tdip_q = (δ_q)*(t-tref_q) # Effective dipolar evolution time
        if fractal:
            Bbasis = dl.bg_strexp(tdip_q,conc_decay*λ,d)
        else:
            Bbasis = dl.bg_hom3d(tdip_q,conc_decay,λ)

    return Bbasis 
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def Bbasis2D(t1,t2,λ,conc_decay,d,δ1,δ2,tref1,tref2,fractal=False):
    
    x = np.newaxis 
    t1 = t1[:,x]
    t2 = t2[x,:]

    Bbasis = 1
    for δ1_q,tref1_q,δ2_q,tref2_q in zip(δ1,tref1,δ2,tref2):
        tdip_q = abs(δ1_q)*(t1-tref1_q) + abs(δ2_q)*(t2-tref2_q) # Effective dipolar evolution time
        if fractal:
            Bbasis = _strexp(tdip_q,conc_decay*λ,d)
        else:
            Bbasis = _hom3d(tdip_q,conc_decay,λ)

    return Bbasis 
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def add_circlularpermutations(varlist,var):
    for _ in range(len(var)):
        # Roll vectors over all spins interactions
        var = np.roll(var,-1,axis=0)
        # Add to list
        varlist.append(var) 
    return varlist
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def add_permutations(varlist,var):
    var = np.atleast_1d(var)
    for perm in set(permutations(np.arange(len(var)))):
        # Add to list
        varlist.append(np.atleast_1d([var[p] for p in perm])) 
    return varlist
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def get_experimental_taus(descriptor):
    def getdelay(delay):
        PulseSPEL = descriptor['DSL']['ftEpr']['PlsSPELGlbTxt']
        # Split lines
        for line in PulseSPEL.split('\\n'):
            if delay in line:
                # Split words
                words = line.split(' ')
                # Get value assigned to delay
                index = [i for i,x in enumerate(words) if x==delay][0]
                index += 1
                while words[index]=='' or words[index]=='=':
                    param = words[index+1]
                    index += 1
                param = float(param)
                break
        return param/1000 # ns->us
    return getdelay('d3'),getdelay('d1'),getdelay('d2')
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def kde2d(x,y,color,bins):
    deltaX = (max(x) - min(x))/10
    deltaY = (max(y) - min(y))/10
    xmin, xmax = min(bins),max(bins)
    ymin, ymax = min(bins),max(bins)
    xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
    positions = np.vstack([xx.ravel(), yy.ravel()])
    values = np.vstack([x, y])
    kernel = st.gaussian_kde(values,bw_method=0.6)
    f = np.reshape(kernel(positions).T, xx.shape)

    f = f/np.sum(f)
    n = 2000
    t = np.linspace(0, f.max(), n)
    integral = ((f >= t[:, None, None]) * f).sum(axis=(1,2))

    from scipy import interpolate
    interp = interpolate.interp1d(integral, t, fill_value='extrapolate')
    t_contours = np.sort(interp(np.array([0.9,0.5,0.25])))
    plt.contourf(xx,yy, f, np.array([t_contours[0],t_contours[1]]), extent=[x.min(),x.max(),x.min(),x.max()], colors=color,alpha=0.2)
    plt.contourf(xx,yy, f, np.array([t_contours[1],t_contours[2]]), extent=[x.min(),x.max(),x.min(),x.max()], colors=color,alpha=0.4)
    plt.contourf(xx,yy, f, np.array([t_contours[2],np.max(f)]), extent=[x.min(),x.max(),x.min(),x.max()], colors=color,alpha=0.7)

    plt.xlim([xmin,xmax])
    plt.ylim([xmin,xmax])
    plt.grid(alpha=0.3)
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def match(v1, v2, dist):
    assert v1.ndim == v2.ndim == 1
    assert v1.shape[0] == v2.shape[0]
    n = v1.shape[0]
    t = np.dtype(dist(v1[0], v2[0]))
    dist_matrix = np.fromiter((dist(x1, x2) for x1 in v1 for x2 in v2),
                              dtype=t, count=n*n).reshape(n, n)
    _, col_ind = optimize.linear_sum_assignment(dist_matrix)
    return col_ind
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def inverse_sample_function(dist, pnts, x_min=-100, x_max=100, n=1e5,**kwargs):
    x = np.linspace(x_min, x_max, int(n))
    cumulative = np.cumsum(dist(x, **kwargs))
    cumulative -= cumulative.min()
    f = interp1d(cumulative/cumulative.max(), x)
    return f(np.random.random(int(pnts)))
# ----------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------
def plot_multivariate(means,cholfactors,model='gauss',xlim=[0,10],MDfile=None,HSCfile=None,MMMfile=None,MMMparam=None,graphic=None,skiprows=None,saveas=None):

    means,cholfactors = np.atleast_1d(means,cholfactors)

    MDbins = np.linspace(*xlim,150)

    # Monte-Carlo multivariate distance integration
    Nsamples = 10000
    np.random.seed(seed=1)
    if model=='gauss':
        rsamples = Pmultivar(means,cholfactors.tolist()).rvs(Nsamples)
        rsamples = np.maximum(rsamples,1e-16) # Avoid values exactly at zero 
    elif model=='wormgauss':
        # Monte-Carlo multivariate distance integration
        Nsamples = 500000
        rsamples = []
        for contour,persistence,std in zip(means, cholfactors):
            P = lambda r: dl.dd_wormgauss(r,contour,persistence,std) 
            rsamples.append( inverse_sample_function(P, Nsamples, x_min=0, x_max=8, n=300) )
        rsamples = np.array(rsamples).T
        rsamples = np.maximum(rsamples,1e-16) # Avoid values exactly at zero 
        rsamples = np.maximum(rsamples,1e-16) # Avoid values exactly at zero 

    # Triangle inequalities
    Ndistances = np.shape(rsamples)[1]
    idx = np.arange(Ndistances)
    triangle_condition = np.full_like(rsamples,False)
    for n in range(Ndistances):
        idx = np.roll(idx,-1,axis=0)
        triangle_condition[:,n] = np.sum(rsamples[:,idx[:-1]],axis=1) > rsamples[:,idx[-1]]
    triangle_condition = np.all(triangle_condition,axis=1)
    # Discard samples that do not satisfy triangle inequalities
    rsamples = rsamples[triangle_condition,:]

    if MDfile is not None:
        MDdata = np.loadtxt(MDfile,skiprows=skiprows,usecols=[11,12,13])
        MDdata = MDdata/10
        MDmeans = np.median(MDdata,axis=0)
        sorting = match(means,MDmeans,lambda x1, x2: abs(x1 - x2))
        MDdata = MDdata[:,sorting]

    if HSCfile is not None: 
        HSCdata = np.loadtxt(HSCfile)
        HSCmeans = np.median(HSCdata,axis=0)
        sorting = match(means,HSCmeans,lambda x1, x2: abs(x1 - x2))
        HSCdata = HSCdata[:,sorting]

    if MMMfile is not None: 
        MMMdata = loadmat(MMMfile)['rlist']
        MMMmeans = np.median(MMMdata,axis=0)
        sorting = match(means,MMMmeans,lambda x1, x2: abs(x1 - x2))
        MMMdata = MMMdata[:,sorting]

    if MMMparam is not None:
        truth_means,truth_std = MMMparam
        truth_means,truth_std =np.atleast_1d(truth_means,truth_std)
        sorting = match(means,truth_means,lambda x1, x2: abs(x1 - x2))
        truth_means = truth_means[sorting]
        truth_std = truth_std[sorting]
        truth_corr = [truth_std[0], truth_std[1], truth_std[2], 0, 0, 0]

    # Define custom colors
    violet = '#4550e6'

    # Get dimensionality of the distribution
    Ndims = len(means)

    # Construct grids for numerical evaluation
    x = np.linspace(*xlim,200)
    grid1d = np.stack(np.meshgrid(x, indexing='ij'),axis=-1)
    grid2d = np.stack(np.meshgrid(x, x, indexing='ij'),axis=-1)

    nrow,ncol = Ndims,Ndims
    fig, axes = plt.subplots(
        nrow, ncol,
        gridspec_kw=dict(wspace=0.05, hspace=0.05,
                        top=1. - 0.5 / (nrow + 1), bottom=0.5 / (nrow + 1),
                        left=0.5 / (ncol + 1), right=1 - 0.5 / (ncol + 1)),
        figsize=(9, 9),
    )
    plt.rcParams['font.size'] = '13'

    if graphic is not None:
        import matplotlib.image as mpimg
        ax = plt.subplot(Ndims,Ndims,Ndims-1)
        img =  mpimg.imread(graphic)
        plt.imshow(img)
        plt.gca().get_xaxis().set_visible(False)
        plt.gca().get_yaxis().set_visible(False)
        plt.gca().spines['top'].set_visible(False)
        plt.gca().spines['bottom'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
        plt.gca().spines['left'].set_visible(False)

    # Loop over all subplots
    for dim1 in range(Ndims):
        for dim2 in range(Ndims):
                
            plt.sca(axes[dim1][dim2])

            if dim2>dim1:
                
                plt.gca().get_xaxis().set_visible(False)
                plt.gca().get_yaxis().set_visible(False)
                plt.gca().spines['top'].set_visible(False)
                plt.gca().spines['bottom'].set_visible(False)
                plt.gca().spines['right'].set_visible(False)
                plt.gca().spines['left'].set_visible(False)
                continue 

            # Univariate marginal distribution
            if dim1==dim2:
                #----------------------------------------------------------------------------------
                # Plot simulations
                if MDfile is not None: 
                    kde = st.gaussian_kde(MDdata[:,dim1])
                    plt.fill(MDbins,kde(MDbins),color='grey',alpha=0.5,density=True,label='MD simulation')
                if HSCfile is not None: 
                    kde = st.gaussian_kde(HSCdata[:,dim1])
                    plt.fill(MDbins,kde(MDbins),color='grey',alpha=0.5,label='HSC simulation')
                if MMMfile is not None: 
                    kde = st.gaussian_kde(MMMdata[:,dim1])
                    plt.fill(MDbins,kde(MDbins),color='grey',alpha=0.5,label='MMM simulation')
                if MMMparam is not None: 
                    Σ = cholesky_covmat(truth_corr)
                    truth_pdf = Punivar(truth_means[dim1],0,Σ[dim1,dim1]).pdf(grid1d)
                    plt.fill(x,truth_pdf,color='grey',alpha=0.3,label='MMM simulation',linewidth=0)
                #----------------------------------------------------------------------------------
                kde = st.gaussian_kde(rsamples[:,dim1])
                plt.fill(MDbins,kde(MDbins),color=violet,alpha=0.7,label='Experiment')

                plt.gca().set_xlim(xlim)
                plt.gca().get_xaxis().set_ticks(np.arange(min(xlim)+1,max(xlim)+1,2).tolist())
                plt.gca().get_xaxis().set_ticklabels(np.arange(min(xlim)+1,max(xlim)+1,2).tolist())
                plt.gca().get_yaxis().set_visible(False)
                plt.gca().spines['top'].set_visible(False)
                plt.gca().spines['right'].set_visible(False)
                plt.gca().spines['left'].set_visible(False)
                if dim1==0:
                    plt.legend(frameon=False)
                if dim1<Ndims-1:
                    plt.gca().get_xaxis().set_ticklabels([])
                else:
                    plt.xlabel(f'$r_{dim2+1}$ (nm)')
                plt.rcParams['axes.ymargin'] = 0

            if dim2<dim1:
                #----------------------------------------------------------------------------------
                # Plot simulations
                if MDfile is not None: 
                    subsampling = np.arange(0,MDdata.shape[0],50)
                    kde2d(MDdata[subsampling,dim2], MDdata[subsampling,dim1],color='grey', bins=MDbins)
                if HSCfile is not None: 
                    subsampling = np.arange(0,HSCdata.shape[0],50)
                    kde2d(HSCdata[subsampling,dim2], HSCdata[subsampling,dim1],color='grey', bins=MDbins)
                if MMMfile is not None: 
                    subsampling = np.arange(0,MMMdata.shape[0],50)
                    kde2d(MMMdata[subsampling,dim2], MMMdata[subsampling,dim1],color='grey', bins=MDbins)
                #----------------------------------------------------------------------------------
                plt.grid(alpha=0.2)
                subsampling = np.arange(0,rsamples.shape[0],50)
                kde2d(rsamples[subsampling,dim2], rsamples[subsampling,dim1],color=violet, bins=MDbins)

                #cont3 = plt.contour(x,x,pdf,alpha=0.7,cmap='binary')
                plt.xlim(xlim)
                plt.ylim(xlim)
                
                plt.gca().get_xaxis().set_ticks(np.arange(min(xlim)+1,max(xlim)+1,2).tolist())
                plt.gca().get_yaxis().set_ticks(np.arange(min(xlim)+1,max(xlim)+1,2).tolist())
                plt.gca().get_xaxis().set_ticklabels(np.arange(min(xlim)+1,max(xlim)+1,2).tolist())
                plt.gca().get_yaxis().set_ticklabels(np.arange(min(xlim)+1,max(xlim)+1,2).tolist())
                if dim2>0:
                    plt.gca().get_yaxis().set_ticklabels([])
                else:
                    plt.ylabel(f'$r_{dim1+1}$ (nm)')
                if dim1<Ndims-1:
                    plt.gca().get_xaxis().set_ticklabels([])
                else:
                    plt.xlabel(f'$r_{dim2+1}$ (nm)')
    
    if saveas is not None:
        plt.savefig(saveas)
    plt.show()
# ----------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------
def plot_dataset_fit(ts,Vexps,model,params,scales,saveas=None,xlim=None):
    nrow,ncol = 1,2
    fig, axes = plt.subplots(
    nrow, ncol,
    gridspec_kw=dict(wspace=0.05, hspace=0.05,
                    top=1. - 0.5 / (nrow + 1), bottom=0.5 / (nrow + 1),
                    left=0.5 / (ncol + 1), right=1 - 0.5 / (ncol + 1)),
                    sharey=True,
    figsize=(ncol*5, nrow*3*len(Vexps))
    )

    for ii,(t,Vexp,param) in enumerate(zip(ts,Vexps,params)):
        Vfit, λ0, Vinter, Vintra,twospin_contributions,threespin_contributions = model(*param)
        Vfit = Vfit*scales[ii] 
        twospin_contrib = 0
        threespin_contrib = 0
        for cont in twospin_contributions: 
            twospin_contrib = twospin_contrib + cont
        for cont in threespin_contributions: 
            threespin_contrib = threespin_contrib + cont

        # Define custom colors
        pastelgreen = '#5fb0b9ff'
        pastelmagenta = '#f06b82fd'
        violet = '#4550e6'

        plt.rcParams['font.size'] = '13'
        plt.rcParams['axes.xmargin'] = 0.025

        axes[0].plot(t,Vexp+ii*3/4,'.',color='grey')
        axes[0].plot(t,Vfit+ii*3/4,'-',color=violet,linewidth=3)
        axes[0].set_ylabel('$V(t)$')
        axes[0].set_xlabel('$t$ (μs)')
        
        axes[1].plot(t,(λ0+twospin_contrib)*Vinter+ii*3/4,color=pastelgreen,linewidth=3)
        axes[1].plot(t,(λ0+threespin_contrib)*Vinter+ii*3/4,color=pastelmagenta,linewidth=3)
        axes[1].set_xlabel('$t$ (μs)')

        if xlim is not None:
            axes[0].set_xlim(xlim)
            axes[1].set_xlim(xlim)


    if saveas is not None:  
            plt.savefig(saveas)
    plt.show()  
# ----------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------
def plot_trier_fit(t1,t2,Vexp,model,param,scale,saveas=None,levels=15,cmap='turbo'):

    Vfit, λ0, Vinter, Vintra,twospin_contributions,threespin_contributions = model(*param)
    Vfit = Vfit*scale
    
    violet = '#4550e6'
    fig = plt.figure(figsize=(10, 9))
    outer = gridspec.GridSpec(2, 2, wspace=0.0, hspace=0.0)
    plt.rcParams['font.size'] = '13'
    plt.rcParams['axes.ymargin'] = 0.1
    plt.rcParams['axes.xmargin'] = 0.1

    ax = plt.Subplot(fig, outer[3])
    ax.contourf(t2,t1,Vexp,levels,cmap=cmap)
    ax.contour(t2,t1,Vfit,levels,cmap='bone_r', linewidths=1.5, linestyles='solid')
    ax.set_xlabel('$t_2$ (μs)')
    ax.set_ylabel('$t_1$ (μs)')
    ax.yaxis.set_label_position("right")
    ax.yaxis.tick_right()
    fig.add_subplot(ax)

    inner = gridspec.GridSpecFromSubplotSpec(1, 3,
                    subplot_spec=outer[2], wspace=0.1, hspace=0.0)

    ax = plt.Subplot(fig, inner[-1])
    ax.plot(np.sum(Vexp,axis=1),t1, '.', color='grey')
    ax.plot(np.sum(Vfit,axis=1),t1, color=violet, linewidth=2.5)
    ax.set_ylabel('$t_1$ (μs)')
    ax.set_xlabel('$V(t)$')
    ax.get_yaxis().set_visible(False)
    ax.get_xaxis().set_visible(False)
    ax.invert_xaxis()
    ax.set_ylim([min(t1),max(t1)])
    fig.add_subplot(ax)

    inner = gridspec.GridSpecFromSubplotSpec(3, 1,
                    subplot_spec=outer[1], wspace=0.1, hspace=0.0)

    ax = plt.Subplot(fig, inner[-1])
    ax.plot(t2, np.sum(Vexp,axis=0), '.', color='grey')
    ax.plot(t2, np.sum(Vfit,axis=0), color=violet, linewidth=2.5)
    ax.set_xlabel('$t_2$ (μs)')
    ax.set_ylabel('$V(t)$')
    ax.get_yaxis().set_visible(False)
    ax.get_xaxis().set_visible(False)
    ax.set_xlim([min(t2),max(t2)])
    fig.add_subplot(ax)


    if saveas is not None:  
            plt.savefig(saveas)
    plt.show()
# ----------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------
def covariance_summary(means,cholesky_factors):
    Σ = cholesky_covmat(cholesky_factors) 
    N = Σ.shape[0]
    sigmas_string,corr_string,means_string = '','',''
    for n in range(N):
        for m in range(N):
            if n == m:
                sigmas_string += f'  σ{n+1} = {np.sqrt(Σ[n,n]):.4f} nm \n'
            else:
                corr_string += f'  ρ{n+1}{m+1} = {Σ[n,m]/np.sqrt(Σ[n,n])/np.sqrt(Σ[m,m]):.4f} \n'
    for n,mean in enumerate(means):
        means_string += f'  rmean{n+1} = {mean} nm \n'

    print('--------------------------------------------------------')
    print('Multivariate Gaussian parameters')
    print('--------------------------------------------------------')
    print('Means:')
    print(means_string)
    print('Standard deviations:')
    print(sigmas_string)
    print('Correlation coefficients:')
    print(corr_string)
# ----------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------
def get_corrs(file):
    if 'txt' in file:
        HSCdata = np.loadtxt(file)
    if 'mat' in file:
        HSCdata = loadmat(file)['rlist']
    corrs = []
    for n1,n2 in [[0,1],[1,2],[2,0]]:
        hist,bins1,bins2 = np.histogram2d(HSCdata[:,n1],HSCdata[:,n2],bins=30)
        bins1 = bins1[:-1] + (bins1[1]-bins1[0])/2 
        bins2 = bins2[:-1] + (bins2[1]-bins2[0])/2
        xx, yy = np.meshgrid(bins1, bins2)
        grid2d = np.empty(xx.shape + (2,))
        grid2d[:, :, 0] = xx
        grid2d[:, :, 1] = yy
        
        def Pmodelflat(mean1,mean2,std1,std2,corr):
            L = np.atleast_2d([
                [std1,    0],
                [corr, std2]])
            P = Pbivar([mean1,mean2],L@L.T).pdf(grid2d)
            return P.flatten() 
        model = dl.Model(Pmodelflat)
        model.mean1.set(par0=3,lb=0,ub=8)
        model.mean2.set(par0=3,lb=0,ub=8)
        model.std1.set(par0=0.2,lb=0.01,ub=1)
        model.std2.set(par0=0.2,lb=0.01,ub=1) 
        model.corr.set(par0=0,lb=-0.5,ub=0.5)

        fit = dl.fit(model,hist.flatten())
        corrs.append(fit.corr) 
    return corrs
# ----------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------
def get_corrs_4spin(file):
    if 'txt' in file:
        HSCdata = np.loadtxt(file)
    if 'mat' in file:
        HSCdata = loadmat(file)['rlist']
    corrs = []
    for n1,n2 in [[0,1],[0,2],[0,3],[0,4],[0,5],[1,2],[1,3],[1,4],[1,5],[2,3],[2,4],[2,5],[3,4],[3,5],[4,5]]:
        hist,bins1,bins2 = np.histogram2d(HSCdata[:,n1],HSCdata[:,n2],bins=30)
        bins1 = bins1[:-1] + (bins1[1]-bins1[0])/2 
        bins2 = bins2[:-1] + (bins2[1]-bins2[0])/2
        xx, yy = np.meshgrid(bins1, bins2)
        grid2d = np.empty(xx.shape + (2,))
        grid2d[:, :, 0] = xx
        grid2d[:, :, 1] = yy
        
        def Pmodelflat(mean1,mean2,std1,std2,corr):
            L = np.atleast_2d([
                [std1,    0],
                [corr, std2]])
            P = Pbivar([mean1,mean2],L@L.T).pdf(grid2d)
            return P.flatten() 
        model = dl.Model(Pmodelflat)
        model.mean1.set(par0=3,lb=0,ub=8)
        model.mean2.set(par0=3,lb=0,ub=8)
        model.std1.set(par0=0.2,lb=0.01,ub=1)
        model.std2.set(par0=0.2,lb=0.01,ub=1) 
        model.corr.set(par0=0,lb=-0.5,ub=0.5)

        fit = dl.fit(model,hist.flatten())
        corrs.append(fit.corr) 
    return corrs
# ----------------------------------------------------------------------------------
